{-# LANGUAGE FlexibleInstances #-}
module Interface where

import Control.Applicative
import Data.Foldable
import Data.Traversable
import qualified Data.Map as M
import Data.Monoid
import Data.Set(Set)
import qualified Data.Set as S

import Label
import Type

type Interface = TypeLike () (Label, Label)

iunit :: Interface
iunit = TVar ()

shape :: TypeLike v a -> TypeLike v ()
shape = unTypeOnConstr . fmap (const ()) . TypeOnConstr

newInterface :: TypeLike v bt -> LabelGen (TypeLike v (Label, Label))
newInterface typ =
  unTypeOnConstr <$> traverse (const ((,) <$> newLabel "q" <*> newLabel "a"))
                            (TypeOnConstr typ)

forgetVars :: TypeLike v bt -> TypeLike () bt
forgetVars = unTypeOnVar . fmap (const ()) . TypeOnVar

initial :: Interface -> Set Label
initial i = case i of
  TConstr (q, _) _ -> S.singleton q
  _ :=>: b     -> initial b
  a :*: b      -> initial a `mappend` initial b
  TVar ()      -> mempty

opponent :: Interface -> Set Label
opponent i = case i of
  TConstr (q, _) _ -> S.singleton q
  a :=>: b     -> player   a `mappend` opponent b
  a :*:  b     -> opponent a `mappend` opponent b
  TVar ()      -> mempty

player :: Interface -> Set Label
player i = case i of
  TConstr (_, a) _ -> S.singleton a
  a :=>: b     -> opponent a `mappend` player b
  a :*:  b     -> player   a `mappend` player b
  TVar ()      -> mempty

qaPairs :: Interface -> Set (Label, Label)
qaPairs = foldMap S.singleton . TypeOnConstr

question, answer :: Interface -> Set Label
question = S.map fst . qaPairs
answer   = S.map snd . qaPairs

unifyInterface :: Interface -> Interface -> LSubst
unifyInterface i1 i2 = case (i1, i2) of
  (TVar (), TVar ())               -> mempty
  (TConstr (q1, a1) _, TConstr (q2, a2) _) -> M.fromList [(q2, q1), (a2, a1)]
  (a1 :=>: b1, a2 :=>: b2)         -> unifyInterface a1 a2
                            `mappend` unifyInterface b1 b2
  (a1 :*:  b1, a2 :*:  b2)         -> unifyInterface a1 a2
                            `mappend` unifyInterface b1 b2
  x                                -> error $ "unifyInterface" ++ show x

instance LabelSubst (TypeLike () (Label, Label)) where
  labelSubst subst i = unTypeOnConstr $ fmap (labelSubst subst) (TypeOnConstr i)

dummy :: Interface -> LabelGen Interface
dummy = newInterface
