{-# LANGUAGE GeneralizedNewtypeDeriving #-}
module Label where

import Control.Applicative
import Control.Monad.State
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe

data Label = Label String Int deriving (Eq, Ord)
type Locus = Maybe String

instance Show Label where
  show (Label s n) = s ++ show n

class MonadLabelGen m where
  newLabel  :: String -> m Label
  withLocus :: Locus -> m a -> m a
  getLocus  :: m Locus

data LabelGenState = LabelGenState
  { lgLabels :: [Int]
  , lgLocus  :: Locus
  }
newtype LabelGen a = LabelGen {unLabelGen :: State LabelGenState a}
  deriving (Applicative, Functor, Monad, MonadState LabelGenState)

instance MonadLabelGen LabelGen where
  newLabel s = do
    l:ls <- gets lgLabels
    modify $ \st -> st {lgLabels = ls}
    return $ Label s l
  getLocus = gets lgLocus
  withLocus l m = do
    before <- getLocus
    modify $ \st -> st {lgLocus = l}
    res <- m
    modify $ \st -> st {lgLocus = before}
    return res

evalLabelGen :: LabelGen a -> a
evalLabelGen = flip evalState (LabelGenState [0..] Nothing) . unLabelGen

type LSubst = Map Label Label

class LabelSubst a where
  labelSubst :: LSubst -> a -> a
instance LabelSubst Label where
  labelSubst subst l = M.findWithDefault l l subst
instance (LabelSubst a, LabelSubst b) => LabelSubst (a, b) where
  labelSubst subst (x, y) = (labelSubst subst x, labelSubst subst y)
instance LabelSubst a => LabelSubst [a] where
  labelSubst subst x = fmap (labelSubst subst) x
instance (Ord k, LabelSubst k, LabelSubst e) => LabelSubst (Map k e) where
  labelSubst s = fmap (labelSubst s) . M.mapKeys (labelSubst s)
