{-# LANGUAGE DeriveDataTypeable, RecordWildCards #-}
module Main where

import Control.Monad
import Paths_gamc
import System.Console.CmdArgs.Implicit
import System.Directory
import System.FilePath

import Backend
import Compiler
import Engine
import ParseToAST
import TypeChecker

-- | Parse, typecheck and run the compiler on the argument
compileTerm :: String -> Net
compileTerm = evalCompiler . typeCheckOuter . parseTerm

-- | Commandline arguments
data Gamc = Gamc
  { inputFile :: FilePath
  , outputDir :: FilePath
  , flags     :: String
  } deriving (Show, Data, Typeable)

defArgs :: Mode (CmdArgs Gamc)
defArgs = cmdArgsMode $ Gamc
  { inputFile = def   &= argPos 0 &= typFile
  , outputDir = "out" &=             typDir      &= help "Directory to generate output in"
  , flags     = def   &=             typ "FLAGS" &= help "Arguments passed to mpicc"
  } &= summary "ICA to GAM compiler"

runCompiler :: Gamc -> IO ()
runCompiler Gamc{..} = do
  prog <- readFile inputFile

  root <- readFile =<< getDataFileName "root.c"
  make <- readFile =<< getDataFileName "Makefile_template"

  let (nodeProgs, dotfile) = compileNet (compileTerm prog) root

  createDirectoryIfMissing True outputDir

  preludePath <- getDataFileName "prelude.h"
  copyFile preludePath (outputDir </> "prelude.h")
  nodes <- forM nodeProgs $ \(node, subprog) -> do
    writeFile (outputDir </> node <.> "c") subprog
    return node
  writeFile (outputDir </> "Makefile") $ createMakefile make flags nodes
  writeFile (outputDir </> "diagram" <.> "dot") $ dotfile

main :: IO ()
main = runCompiler =<< cmdArgsRun defArgs
