{-# LANGUAGE FlexibleInstances, TypeOperators #-}
module Pretty where

import Data.List

import AST
import Instr
import Label
import Type

class Pretty a where
  pretty :: a -> String

instance Pretty () where
  pretty () = "()"

instance (Pretty a, Pretty b) => Pretty (a, b) where
  pretty (a,b) = "(" ++ pretty a  ++ ", " ++ pretty b ++ ")"

instance Pretty [Char] where
  pretty v = v

instance Pretty TConstr where
  pretty bt = case bt of
    TInt   -> "Int"
    TFloat -> "Float"
    TCom   -> "Com"
    TArray -> "[]"

instance (Pretty v, Pretty bt) => Pretty (TypeLike v bt) where
  pretty typ = case typ of
    TConstr c [] -> pretty c
    TConstr c ts -> pretty c ++ " " ++ unwords (map pretty ts)
    t1 :=>: t2 -> "(" ++ pretty t1 ++ " -> " ++ pretty t2 ++ ")"
    t1 :*: t2  -> "(" ++ pretty t1 ++ " * "  ++ pretty t2 ++ ")"
    TVar v     -> pretty v

instance Pretty Label where
  pretty = show

instance Pretty term => Pretty (BaseTerm term) where
  pretty term = case term of
    Var v -> pretty v
    Abs v Nothing t -> "(\\" ++ pretty v ++ ". " ++ pretty t ++ ")"
    Abs v (Just typ) t -> "(\\(" ++ pretty v ++ " : "  ++ pretty typ ++ "). " ++ pretty t ++ ")"
    t1 :$ t2 -> "(" ++ pretty t1 ++ " " ++ pretty t2 ++ ")"
    Lit n -> pretty n
    Const c -> cName c
    Pair    -> "(,)"
    Proj1   -> "fst"
    Proj2   -> "snd"
    Fix     -> "fix"
    t :@ n -> pretty t ++ "@" ++ pretty n

instance Pretty a => Pretty (a ::: b) where
  pretty (a ::: _) = pretty a

instance Pretty Term where
  pretty (Term bt) = pretty bt

instance Pretty TypedTerm where
  pretty (TypedTerm t) = pretty t

instance Pretty PrimLit where
  pretty (LInt n) = show n
  pretty (LFloat i1 i2) = show i1 ++ "." ++ show i2

instance Pretty Lit where
  pretty (PrimLit p) = pretty p
  pretty (LArray ls) = "[" ++ intercalate ", " (map pretty ls) ++ "]"
