module Unify where

import Data.Foldable as F
import qualified Data.Map as M
import Data.Monoid

import Pretty
import Type

occurs :: TVar -> Type -> Bool
occurs v = F.any (== v) . TypeOnVar

unify :: Type -> Type -> TSubst
unify typ1 typ2 = case (typ1, typ2) of
  (TConstr c1 ts1, TConstr c2 ts2)
    | c1 == c2 && length ts1 == length ts2 -> mconcat $ zipWith unify ts1 ts2
  (t1 :=>: t2, s1 :=>: s2)    -> u2 `mappend` u1
    where u1 = unify t1 s1
          u2 = unify (typeSubst u1 t2) (typeSubst u1 s2)
  (t1 :*: t2, s1 :*: s2)     -> u2 `mappend` u1
    where u1 = unify t1 s1
          u2 = unify (typeSubst u1 t2) (typeSubst u1 s2)
  (TVar v, TVar v') | v == v' -> mempty
  (TVar v, t)                 -> varCase v t
  (t, TVar v)                 -> varCase v t
  x                           -> error $ "can't unify " ++ pretty x
  where
    varCase v t = if occurs v t
      then error $ "unification occurs check: " ++ v ++ " occurs in " ++ pretty t
      else M.singleton v t
