{-# LANGUAGE TemplateHaskell #-}
module ParseToAST where
import Control.Arrow
import Data.Monoid
import Data.Set(Set)
import qualified Data.Set as S
import Data.Parser.Grempa.Static
import AST
import qualified Lexer
import Parser

typeToType :: PType -> Type
typeToType PTInt = TInt
typeToType (PTFun t1 t2) = typeToType t1 :->: typeToType t2

exprToTerm :: Expr -> Term
exprToTerm pexpr = Term $ case pexpr of
  EApp e1 e2     -> exprToTerm e1 :$ exprToTerm e2
  EOp  "+"       -> Plus
  EOp  "*"       -> Times
  EOp  "-"       -> Minus
  EVar v         -> Var v
  ENum i         -> Lit $ fromIntegral i
  ELam vs e1     -> unTerm $ foldr (\(v, t) -> Term . Abs v t) (exprToTerm e1) (map (second (fmap typeToType)) vs)
  EAt e1 node    -> exprToTerm e1 :@ node
  EFix           -> Fix
  --ECase e brs def -> Case (exprToTerm e) (map (fromIntegral *** exprToTerm) brs) $ exprToTerm def
  EIf            -> If


parseTerm :: String -> Term
parseTerm = exprToTerm . parse $(mkStaticParser fun [|fun|]) . Lexer.lexToks
