#ifndef __PRELUDE_H__
#define __PRELUDE_H__

#include "mpi.h"
// #include <gc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>

#define DATA_EMPTY 0
#define DATA_INTEGER 1
#define DATA_INL 2
#define DATA_INR 3
#define DATA_PAIR 4
#define STACK_SIZE 1000000
#define MALLOC malloc // GC_malloc
#define FREE(x)
#define MAX_MSG_LENGTH STACK_SIZE
// #define PRINT printf("%s", __FUNCTION__); print_data(data)
#define PRINT

void print_data();

/* Data ---------------------------------------------------------------------*/
struct Data_t;
typedef struct Data_t Data;

struct Data_t {
  int   tag;
  struct {
    Data* sum;
    struct {
      Data* fst;
      Data* snd;
    } pair;
    int integer;
  } data;
};

Data  data;
Data* stack;
int   stacktop;

int* intstack;
int  intstacktop;

inline Data Pair(Data fst, Data snd) {
  PRINT;
  Data res;
  res.tag = DATA_PAIR;
  res.data.pair.fst = (Data*)MALLOC(sizeof(Data));
  res.data.pair.snd = (Data*)MALLOC(sizeof(Data));
  *res.data.pair.fst = fst;
  *res.data.pair.snd = snd;
  return res;
}

inline Data Inl(Data x) {
  PRINT;
  Data res;
  res.tag      = DATA_INL;
  res.data.sum = (Data*)MALLOC(sizeof(Data));
  *res.data.sum = x;
  return res;
}

inline Data Inr(Data x) {
  PRINT;
  Data res;
  res.tag      = DATA_INR;
  res.data.sum = (Data*)MALLOC(sizeof(Data));
  *res.data.sum = x;
  return res;
}

inline Data Integer(int n) {
  PRINT;
  Data res;
  res.tag          = DATA_INTEGER;
  res.data.integer = n;
  return res;
}

inline Data Empty() {
  PRINT;
  Data res;
  res.tag         = DATA_EMPTY;
  return res;
}

inline void free_data(Data d) {
  switch (d.tag) {
    case DATA_EMPTY:
      return;
    case DATA_INTEGER:
      return;
    case DATA_INL:
    case DATA_INR:
      free_data(*d.data.sum);
      FREE(d.data.sum);
      return;
    case DATA_PAIR:
      free_data(*d.data.pair.fst);
      free_data(*d.data.pair.snd);
      FREE(d.data.pair.fst);
      FREE(d.data.pair.snd);
      return;
    default:
      return;
  }
}

inline int* serialise(Data d, int* res) {
  *(res++) = d.tag;
  switch (d.tag) {
    case DATA_EMPTY:
      return res;
    case DATA_INTEGER:
      *(res++) = d.data.integer;
      return res;
    case DATA_INL:
    case DATA_INR:
      return serialise(*d.data.sum, res);
    case DATA_PAIR: {
      int* res2 =  serialise(*d.data.pair.fst, res);
      return serialise(*d.data.pair.snd, res2);
    }
    default:
      return res;
  }
}

Data deserialisep(int* input, int** inputpos) {
  switch (*(input++)) {
    case DATA_EMPTY:
      *inputpos = input;
      return Empty();
    case DATA_INTEGER:
      *inputpos = input + 1;
      return Integer(*input);
    case DATA_INL:
      return Inl(deserialisep(input, inputpos));
    case DATA_INR:
      return Inr(deserialisep(input, inputpos));
    case DATA_PAIR: {
      Data fst = deserialisep(input,     inputpos);
      Data snd = deserialisep(*inputpos, inputpos);
      return Pair(fst, snd);
    }
  }
}

inline Data deserialise(int* input) {
  PRINT;
  int* inputpos = input;
  return deserialisep(input, &inputpos);
}

inline void initialise(int argc, char** argv) {
  // GC_init();
  MPI_Init(&argc, &argv);
  data = Empty();
  stack = (Data*)malloc(sizeof(Data) * STACK_SIZE);
  intstack = (int*)malloc(sizeof(int) * STACK_SIZE);
  stacktop = 0;
  intstacktop = 0;
}

inline void deinitialise() {
  free(stack);
  free(intstack);
}

inline void push_stack(Data d) {
  PRINT;
  stack[stacktop] = d;
  // printf("push stack:"); print_data(stack[stacktop]);
  ++stacktop;
}

inline Data pop_stack() {
  PRINT;
  --stacktop;
  // printf("pop stack stack top:"); print_data(stack[stacktop]);
  return stack[stacktop];
}

inline void push_int_stack(int n) {
  PRINT;
  intstack[intstacktop] = n;
  ++intstacktop;
}

inline int pop_int_stack() {
  PRINT;
  --intstacktop;
  return intstack[intstacktop];
}

/* Instructions -------------------------------------------------------------*/

inline void inl() {
  PRINT;
  data = Inl(data);
}

inline void inr() {
  PRINT;
  data = Inr(data);
}

inline void fst() {
  PRINT;
  Data temp1 = *data.data.pair.fst;
  Data temp2 = *data.data.pair.snd;
  FREE(data.data.pair.fst);
  FREE(data.data.pair.snd);
  push_stack(temp2);
  data = temp1;
  PRINT;
}

inline void snd() {
  PRINT;
  Data temp1 = *data.data.pair.fst;
  Data temp2 = *data.data.pair.snd;
  FREE(data.data.pair.fst);
  FREE(data.data.pair.snd);
  push_stack(temp1);
  data = temp2;
}

inline void unfst() {
  PRINT;
  data = Pair(data, pop_stack());
}

inline void unsnd() {
  PRINT;
  data = Pair(pop_stack(), data);
}

inline void flip() {
  PRINT;
  Data d1 = pop_stack();
  Data d2 = pop_stack();
  push_stack(d1);
  push_stack(d2);
}

inline void push() {
  PRINT;
  push_stack(data);
  data = Empty();
}

inline void pop() {
  PRINT;
  Data d = pop_stack();
  free_data(d);
}

inline void integer(int n) {
  PRINT;
  free_data(data);
  data = Integer(n);
}

inline void pushint() {
  push_int_stack(data.data.integer);
  data = Empty();
}
/*
inline void bang() {
  PRINT;
  integer(0);
}
*/

inline void add() {
  PRINT;
  int n1 = pop_int_stack();
  // printf("add n1: %d\n", n1);
  int n2 = data.data.integer;
  // printf("add n2: %d\n", n2);
  integer(n1 + n2);
}

inline void sub() {
  PRINT;
  int n1 = pop_int_stack();
  // printf("add n1: %d\n", n1);
  int n2 = data.data.integer;
  // printf("add n2: %d\n", n2);
  integer(n1 - n2);
}

inline void mul() {
  PRINT;
  int n1 = pop_int_stack();
  // printf("add n1: %d\n", n1);
  int n2 = data.data.integer;
  // printf("add n2: %d\n", n2);
  integer(n1 * n2);
}

inline void match(void (*left)(), void (*right)()) {
  PRINT;
  int tag = data.tag;
  // printf("match tag %d\n", tag);
  Data temp = *data.data.sum;
  FREE(data.data.sum);
  data = temp;
  switch (tag) {
    case DATA_INL:
      left();
      break;
    case DATA_INR:
      right();
      break;
    default:
      break;
  }
}

inline void print() {
  PRINT;
  printf("\n##########################");
  printf("\nResult: %d\n", data.data.integer);
}

inline void done() {
  PRINT;
  free_data(data);
  data = Empty();
}

inline void send(int node, int port) {
  PRINT;
  int data_flat[MAX_MSG_LENGTH];
  int* end = serialise(data, data_flat);
  int i;
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);  /* get current process id */

  printf("send from %d to %d port: %d\n", myid, node, port);
  print_data(data);
  ptrdiff_t size = (int)(end - data_flat);
  MPI_Send(&data_flat, size, MPI_INT, node, port, MPI_COMM_WORLD);
  free_data(data);
  data = Empty();
}

inline void print_data(Data data) {
  int data_flat[MAX_MSG_LENGTH];
  int* end = serialise(data, data_flat);
  int i;
  printf("(");
  for(i = 0; &data_flat[i] < end; ++i)
    printf("%d,", data_flat[i]);
  printf(")\n");
  fflush(0);
}

inline int receive() {
  PRINT;
  int data_flat[MAX_MSG_LENGTH];
  MPI_Status stat;
  MPI_Recv(&data_flat, MAX_MSG_LENGTH, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);  /* get current process id */
  printf("receive on %d port: %d\n", myid, stat.MPI_TAG);
  data = deserialise(data_flat);
  print_data(data);
  return stat.MPI_TAG;
}

#endif
